/*
 * imagemanager2.cpp
 *
 *  Created on: 26-02-2013
 *      Author: crm
 */

#include "imagemanager.h"

#include <map>
#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include <GL/glee.h>

#include "resource.h"
#include "log.h"

#include "image.h"
#include "../render.h"

namespace Graphics
	{
	namespace ImageManager
		{
		std::map<std::string, Image *> imgs;
		}

	namespace ImageManager
		{
		void clear(const std::string& name, Image *img);
		void reload(const std::string& name, Image *img);

		bool loadGfx(const std::string& name, Image *img);
		bool loadGfx(const std::string& name, Image *img, unsigned char *pixs, Type type);
		bool loadTex(const std::string& name, Image *img);

		const Image* addImage(Image *img);
		}
	}

using namespace Graphics;

/**** Zewnetrzne ****/
void ImageManager::reload()
	{
	Log::debug("ImageManager.clear: Ponowne wczytywanie %d obrazków", imgs.size());

	for(std::map<std::string, Image *>::iterator it=imgs.begin(); it!=imgs.end(); ++it)
		reload(it->first, it->second);
	}

void ImageManager::flush()
	{
	for(std::map<std::string, Image *>::iterator it=imgs.begin(); it!=imgs.end(); ++it)
		{
		if(it->second->refs()>0)
			continue;

		Log::debug("ImageManager.flush: Kasowanie \"%s\"", it->first.c_str());

		clear(it->first, it->second);
		imgs.erase(it--);
		}
	}

void ImageManager::clear()
	{
	if(imgs.size()<1)
		return;

	Log::info("ImageManager.clear: Czyszczenie, %d obrazków", imgs.size());

	for(std::map<std::string, Image *>::iterator it=imgs.begin(); it!=imgs.end(); ++it)
		{
		if(it->second->refs()>0)
			{
			Log::debug("ImageManager.clear: Obrazek \"%s\" ma referencji: %d", it->first.c_str(), it->second->refs());
			continue;
			}

		clear(it->first, it->second);
		imgs.erase(it--);
		}

	//imgs.clear();

	Log::ok("ImageManager.clear: Wyczyszczono");
	}

/**** Wewnetrzne ****/
void ImageManager::reload(const std::string& name, Image *img)
	{
	if(img->srf)
		loadTex(name, img);		// Potrzebne mi tu sprawdzanie?
	else
		{
		loadGfx(name, img);
		loadTex(name, img);
		}
	}

void ImageManager::clear(const std::string& name, Image *img)
	{
	if(img->tex)
		glDeleteTextures(1, img->tex);
	if(img->srf)
		SDL_FreeSurface(img->srf);

	delete img;
	}


const Image* ImageManager::get(const std::string& name)
	{
	Image *img;

	if(imgs.find(name)!=imgs.end())
		{
		img=imgs[name];
		//img->ref();
		return img;
		}

	try
		{
		img=new Image(0u, 0, 0, NULL);
		}
	catch(std::bad_alloc &)
		{
		Log::fail("ImageManager.load: Nie udalo sie zajac miejsca na \"%s\"", name.c_str());
		return NULL;
		}

	if(!loadGfx(name, img))
		{
		Log::fail("ImageManager.load: Nie udalo sie wczytac grafiki \"%s\"", name.c_str());
		delete img;
		return NULL;
		}

	if(!loadTex(name, img))
		{
		Log::fail("ImageManager.load: Nie udalo sie wygenerowac tekstury \"%s\"", name.c_str());
		delete img;
		return NULL;
		}

	if(img->srf)
		{
		SDL_FreeSurface(img->srf);
		img->srf=NULL;
		}

	Log::debug("ImageManager.load: Wczytano \"%s\"", name.c_str());

	//img->ref();
	imgs[name]=img;

	return img;
	}

const Image* ImageManager::get(const std::string& name, SDL_Surface *srf)
	{
	Image *img;

	if(imgs.find(name)!=imgs.end())
		{
		img=imgs[name];
		//img->ref();
		return img;
		}

	if(!srf)
		return NULL;

	Log::debug("ImageManger.load(srf): Wczytywanie \"%s\"", name.c_str());

	try
		{
		img=new Image(0u, 0, 0, NULL);	// ustawiane nizej~
		}
	catch(std::bad_alloc&)
		{
		Log::fail("ImageManager.load(srf): Nie udalo sie zajac pamieci na \"%s\"", name.c_str());
		return NULL;
		}

	img->srf=SDL_DisplayFormatAlpha(srf);
	img->w=srf->w;
	img->h=srf->h;

	if(!loadTex(name, img))
		{
		Log::fail("ImageManager.load: Nie udalo sie wygenerowac tekstury \"%s\"", name.c_str());
		delete img;
		return NULL;
		}

	//img->ref();
	imgs[name]=img;

	return img;
	}

const Image* ImageManager::get(const std::string& name, unsigned char *pixs, int w, int h, Type type)
	{
	Image *img;

	if(imgs.find(name)!=imgs.end())
		{
		img=imgs[name];
		//img->ref();
		return img;
		}

	if(!pixs || w<0 || h<0)
		return NULL;

	Log::debug("ImageManger.load(pixs): Wczytywanie \"%s\"", name.c_str());

	try
		{
		img=new Image(0u, w, h, NULL);
		}
	catch(std::bad_alloc&)
		{
		Log::fail("ImageManager.load(pixs): Nie udalo sie zajac pamieci na \"%s\"", name.c_str());
		return NULL;
		}

	if(!loadGfx(name, img, pixs, type))
		{
		Log::fail("ImageManager.load(pixs): Nie udalo sie wygenerowac bitmapy \"%s\"", name.c_str());
		delete img;
		return NULL;
		}

	if(!loadTex(name, img))
		{
		Log::fail("ImageManager.load(pixs): Nie udalo sie wygenerowac tekstury \"%s\"", name.c_str());
		delete img;
		return NULL;
		}

	//img->ref();
	imgs[name]=img;

	return img;
	}

void ImageManager::free(const Image* img)
	{
	for(std::map<std::string, Image *>::iterator it=imgs.begin(); it!=imgs.end(); ++it)
		{
		if(it->second!=img)
			continue;

		//it->second->unref();
		if(it->second->refs()<=0)
			{
			Log::debug("ImageManager.free: Kasowanie \"%s\"", it->first.c_str());

			clear(it->first, it->second);
			imgs.erase(it);
			}

		break;
		}
	}


bool ImageManager::loadGfx(const std::string& name, Image *img)
	{
// Wczytywanie grafiki
	SDL_Surface *tmp;
	char *resGfx=Resource::load(name);
	if(resGfx==NULL)
		{
		Log::fail("ImageManager.loadGfx: Nie udalo sie wczytac grafiki \"%s\"", name.c_str());
		return false;
		}

// Dekodowanie
	tmp=IMG_Load_RW(SDL_RWFromMem(resGfx, Resource::getSize()), 1);    // Ustawianie gfx do sprite
	delete [] resGfx;
	if(tmp==NULL)
		{
		Log::fail("ImageManager.loadGfx: Nie udalo sie zdekompresowac grafiki \"%s\"\n\terr: %s", name.c_str(), IMG_GetError());
		return false;
		}

// Zmiana formatu
	SDL_Surface *spr=SDL_DisplayFormatAlpha(tmp);
	SDL_FreeSurface(tmp);
	if(spr==NULL)
		{
		Log::fail("ImageManager.loadGfx: Nie udalo sie przekonwertowac grafiki \"%s\"", name.c_str());
		return false;
		}

	img->srf=spr;
	img->w=spr->w;
	img->h=spr->h;

	return true;
	}

bool ImageManager::loadGfx(const std::string& name, Image *img, unsigned char *pixs, Type type)
	{
	Log::debug("ImageManager.loadGfx(pixs): NOPE");
	return false;
	}

bool ImageManager::loadTex(const std::string& name, Image *img)
	{
	int err;

	img->tex[0]=0u;
	Render::bind(0u);	// potrzebne?	// Tak. Moze zdarzyc sie, ze nowa tekstura bedzie miala id juz skasowanej i zbindowanej. Praktyka tego dowiodla~

	glGenTextures(1, img->tex);

	err=glGetError();
	if(!img->tex[0] || err!=GL_NO_ERROR)
		{
		Log::fail("ImageManager.loadTex: Nie udalo sie dostac tekstury dla \"%s\"", name.c_str());
		return false;
		}

	//glBindTexture(GL_TEXTURE_2D, img.tex[0]);
	Render::bind(img->tex[0]);

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, img->w, img->h, 0, GL_BGRA, GL_UNSIGNED_BYTE, img->srf->pixels);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);	// GL_LINEAR
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);		// Tylko jesli wspolrzedne sa wieksze od 1 -> profit!... chyba.
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);		// Zobaczymy jak bedzie juz cos widac.

	err=glGetError();
	if(err!=GL_NO_ERROR)
		{
		Log::fail("ImageManager.loadTex: Nie udalo sie wygenerowac tekstury dla \"%s\"\n\terr: %04X (%d)", name.c_str(), err, err);
		return false;
		}

	return true;
	}

