/*
 * spritemanager.cpp
 *
 *  Created on: 25-12-2012
 *      Author: crm
 */

#include "spritemanager.h"

#include <map>
#include <tinyxml.h>

#include "parser.h"
#include "resource.h"
#include "log.h"

#include "sprite.h"
#include "imagemanager.h"
#include "image.h"

/******************************************************************************/
/******************************************************************************/

using namespace Graphics;		// Ide na latwizne~

/**** Funkcje ****/
namespace Graphics
	{
	namespace SpriteManager
		{
	// Funkcje pomocnicze wczytywania
		Sprite* loadImg(const std::string& name);		// Wczytanie i obsluga czystego obrazka
		Sprite* loadTxt(const std::string& name);		// Wczytanie i obsluga klasycznej wersji, z pliku tekstowego
		Sprite* loadXml(const std::string& name);		// Wczytanie i obsluga sprite w xml'u

	// Funkcje pomocnicze txt'ka
		bool txtParseAnim(Sprite *spr, const std::string& name);

	// Funkcje pomocnicze xml'a
		void xmlParseSprite(Sprite *spr, TiXmlElement *elem);
		void xmlParseAnim(Animation *ani, TiXmlElement *elem);
		void xmlParseDir(Direction *dir, TiXmlElement *elem);
		void xmlParseFrame(Frame *f, TiXmlElement *elem);
		}

	/**** Zmienne ****/
	namespace SpriteManager
		{
		std::map<std::string, Sprite *> sprites;		// Buforek spritow
		}

	/**** Stale ****/
	namespace SpriteManager
		{
		const char XML_TAG_SPRITE[]=	"sprite";
		const char XML_TAG_IMAGE[]=		"image";
		const char XML_TAG_ANIM[]=		"anim";
		const char XML_TAG_DIR[]=		"dir";
		const char XML_TAG_FRAME[]=		"frame";
		const char XML_ATR_NAME[]=		"name";
		const char XML_ATR_RELATIVE[]=	"relative";
		const char XML_ATR_SPD[]=		"spd";
		const char XML_ATR_RET[]=		"ret";
		const char XML_ATR_X[]=			"x";
		const char XML_ATR_Y[]=			"y";
		const char XML_ATR_W[]=			"w";
		const char XML_ATR_H[]=			"h";
		const char XML_ATR_SPOTX[]=		"sx";
		const char XML_ATR_SPOTY[]=		"sy";
		const char XML_ATR_ACTX[]=		"ax";
		const char XML_ATR_ACTY[]=		"ay";
		const char XML_ATR_DIRECTION[]=	"value";
		const char XML_ATR_DEFAULT[]=	"default";

		const char XML_DIR_RIGHT[]=		"right";
		const char XML_DIR_RIGHT_DOWN[]="right-down";
		const char XML_DIR_DOWN[]=		"down";
		const char XML_DIR_LEFT_DOWN[]=	"left-down";
		const char XML_DIR_LEFT[]=		"left";
		const char XML_DIR_LEFT_UP[]=	"left-up";
		const char XML_DIR_UP[]=		"up";
		const char XML_DIR_RIGHT_UP[]=	"right-up";
		}
	}


using namespace Graphics;

/******************************************************************************/
/******************************************************************************/
/**** Zestaw standardowy ****/
void SpriteManager::clear()
	{
	if(sprites.size()<1)
		return;

	Log::info("SpriteManager.clear: Czyszczenie, %u sprite'ow", sprites.size());

	for(std::map<std::string, Sprite *>::iterator it=sprites.begin(); it!=sprites.end(); ++it)
		{
		if(it->second->refs()>0)
			{
			Log::debug("SpriteManager.clear: Sprite \"%s\" ma referencji: %d", it->first.c_str(), it->second->refs());
			continue;
			}

		delete it->second;
		}

	sprites.clear();

	Log::ok("SpriteManager.clear: Wyczyszczono");
	}

void SpriteManager::flush()
	{
	Log::info("SpriteManager.flush: Czyszczenie, %u sprite'ow", sprites.size());

	for(std::map<std::string, Sprite *>::iterator it=sprites.begin(); it!=sprites.end(); ++it)
		{
		Sprite *spr=it->second;
		if(spr->refs()>0)
			continue;

		Log::debug("SpriteManager.flush: Kasowanie sprite \"%s\", refs %d", it->first.c_str(), spr->refs());

		delete spr;
		}
	}


/**** Wczytywanie, czesc glowna ***/
Sprite* SpriteManager::get(const std::string& name)
	{
	if(name.empty())
		return NULL;

	if(sprites.find(name)!=sprites.end())
		return sprites[name];

	Parser pars(name, ".");

	Sprite *spr;

	if(pars[pars.count()-1]=="xml")
		spr=loadXml(name);
	else if(pars[pars.count()-1]=="spr")
		spr=loadTxt(name);
	else
		spr=loadImg(name);

	if(!spr)
		Log::fail("SpriteManager.get: Nie udalo sie wczytac \"%s\"", name.c_str());
	else
		{
		if(spr->getAnims()<1)
			Log::fail("SpriteManager.get: Sprite \"%s\" nie ma animacji", name.c_str());

		for(unsigned i=0u; i<spr->getAnims(); ++i)
			{
			//if(spr->getAnim(i).getDir(DEFAULT)==NULL)
			//	Log::fail("SpriteManager.get: Sprite \"%s\" animacja %u nie ma kierunkow", name.c_str(), i);
			if(spr->getAnim(i).getDir(DEFAULT).getFrames()<1)
				Log::fail("SpriteManager.get: Sprite \"%s\" animacja %u nie ma klatek w domyslnej animacji", name.c_str(), i);
			}
		}

	//spr->ref();

	return spr;
	}

void SpriteManager::free(const Sprite *spr)
	{
	spr->unref();

	if(spr->refs()<=0)
		{
		for(std::map<std::string, Sprite *>::iterator it=sprites.begin(); it!=sprites.end(); ++it)
			if(it->second==spr)
				{
				sprites.erase(it);
				break;
				}

		delete spr;
		}
	}


/******************************************************************************/
/******************************************************************************/
/**** Wczytywanie samego obrazka ****/
Sprite* SpriteManager::loadImg(const std::string& name)	// Pozwole sobie zaliczyc wczytywanie obrazka jako 'klasyczne'
	{
	Sprite *spr;
	Animation *ani;
	Direction *dir;
	Frame f(0, 0, 0, 0, 0, 0, 0, 0);

	if(!Resource::find(name))
		{
		Log::fail("SpriteManager.loadImg: Nie znaleziono pliku \"%s\"", name.c_str());
		return NULL;
		}

	try
		{
		spr=new Sprite();
		}
	catch(std::bad_alloc&)
		{
		Log::fail("SpriteManager.loadImg: Nie udalo sie zajac miejsca na sprite \"%s\"", name.c_str());
		return NULL;
		}

	const Image *img=ImageManager::get(name.c_str());

	if(!img)
		{
		Log::fail("SpriteManager.loadImg: Nie udalo sie wczytac grafiki \"%s\"", name.c_str());
		delete spr;
		return NULL;
		}

	try
		{
		ani=new Animation();
		dir=new Direction();
		}
	catch(std::bad_alloc&)
		{
		Log::fail("SpriteManaget.loadImg: Nie udalo sie zajac miejsca na kierunek/animacje \"%s\"", name.c_str());
		delete spr;		// Zdaje sobie sprawe, ze przydaloby sie sprawdzic czy ani jest stworzone - wycieki i te sprawy~
		return NULL;
		}

	spr->setImage(img);

	f.x=0;		f.y=0;
	f.w=img->w;	f.h=img->h;
	f.spotx=0;	f.spoty=0;
	f.actx=0;	f.acty=0;
	dir->addFrame(f);
	ani->addDirection(DEFAULT, dir);
	spr->addAnimation(ani);

	if(sprites.find(name)!=sprites.end())
		delete sprites[name];
	sprites[name]=spr;

	return spr;
	}


Sprite* SpriteManager::loadTxt(const std::string& name)
	{
	// Jak tu weszlo, to jest jakas nazwa.

	Parser pars("", ";");
	Sprite *spr=NULL;

	std::string namegfx=name;
	std::string nameani=name;
	char *list=Resource::load(name);
	if(list)
		{
		pars=list;
		delete [] list;

		if(pars.count()<3)
			Log::debug("SpriteManager.loadTxt: Brak wymaganych informacji o sprite \"%s\"", name.c_str());
		else
			{
			namegfx=pars[0];
			nameani=pars[1];
			}
		}

	try
		{
		spr=new Sprite();
		}
	catch(std::bad_alloc&)
		{
		Log::fail("SpriteManager.loadTxt: Nie udalo sie zajac miejsca na sprite \"%s\"", name.c_str());
		return NULL;
		}

	spr->setImage( ImageManager::get(namegfx) );

	if(!spr->getImg())
		{
		Log::fail("SpriteManager.loadTxt: Nie udalo sie wczytac obrazka \"%s\"", name.c_str());
		//delete spr;
		//return NULL;

		Direction *dir=new Direction();
		Animation *ani=new Animation();

		dir->addFrame(Frame(0, 0, 0, 0, 0, 0, 0, 0));
		ani->addDirection(DEFAULT, dir);
		spr->addAnimation(ani);

		sprites[name]=spr;
		return spr;
		}

	if(!txtParseAnim(spr, nameani))
		{
		Log::fail("SpriteManager.loadTxt: Nie udalo sie wczytac animacji \"%s\":\"%s\"", name.c_str(), pars[1].c_str());
		delete spr;
		return NULL;
		}

	sprites[name]=spr;

	return spr;
	}

bool SpriteManager::txtParseAnim(Sprite *spr, const std::string& name)
	{
	if(name.length()<2)
		return 0;

	/*int w, h;
	w=spr->getImg()->w;
	h=spr->getImg()->h;*/

	Parser parsLine;
	Parser parsSpace;
	parsLine.setDelimiter("\n\r");
	parsSpace.setDelimiter(" \t");

	char *tmp=Resource::load(name);
	if(tmp==NULL)
		{
		Log::debug("SpriteManager.txtParseAnim: Nie znaleziono informacji o animacjach, uzywanie calego obrazka");
		Log::debug("SpriteManager.txtParseAnim: Nope, todo");

		return false;	// true;
		}

	std::string aniname="";
	Animation *ani=NULL;
	Direction *dir=NULL;

	parsLine=tmp;
	delete [] tmp;
	for(int i=0; i<parsLine.count(); i++)
		{
		parsSpace=parsLine[i];

		if(parsSpace.count()<1)
			continue;
		else if(parsSpace[0]=="--")
			continue;
		else if(parsSpace[0]=="anim")
			{
			if(parsSpace.count()<2)	// Wstawianie poklatkowe
				{
				Log::fail("SpriteManager.txtParseAnim: Brak wymaganych informacji (\"%s\")", parsSpace.getText().c_str());
				break;
				}

			if(ani)
				{
				spr->addAnimation(ani, aniname);

				aniname="";
				ani=NULL;
				dir=NULL;
				}

			try
				{
				ani=new Animation();
				dir=new Direction();
				}
			catch(std::bad_alloc &)
				{
				Log::fail("SpriteManager.txtParseAnim: Nie udalo sie zajac miejsca na animacje/kierunek \"%s\"", name.c_str());
				break;
				}

			ani->addDirection(DEFAULT, dir);
			dir->setSpeed( parsSpace.valFloat(1) );
			dir->setReturn( parsSpace.val(2) );

			if(parsSpace.count()>3)
				aniname=parsSpace[3];
			}
		else if(parsSpace[0]=="frame")
			{
			int count=parsSpace.val(1);
			Frame f( parsSpace.val(2), parsSpace.val(3), parsSpace.val(4), parsSpace.val(5) );

			if(parsSpace.count()>7)
				{
				f.spotx=parsSpace.val(6);
				f.spoty=parsSpace.val(7);
				}
			if(parsSpace.count()>9)
				{
				f.actx=parsSpace.val(8);
				f.acty=parsSpace.val(9);
				}

			for(int i=0; i<count; ++i)
				dir->addFrame(f);
			}
		//else if(parsSpace[0]=="anim")
		else if(parsSpace.isVal(0) && parsSpace.count()>6)			// Animacje 'na szybko'	// Proste zalozenie, jesli pierwsza wartosc jest liczba to dodawana jest animacja
			{
			// 0  1  2  3  4  5  6
			// c  x  y  w  h sx sy
			// 5  0 24 12 24  7 24

			//printf("%s", parsSpace.getText().c_str());

			if(ani)
				{
				spr->addAnimation(ani, aniname);

				aniname="";
				ani=NULL;
				dir=NULL;
				}

			try
				{
				ani=new Animation();
				dir=new Direction();
				}
			catch(std::bad_alloc&)
				{
				Log::fail("SpriteManager.txtParseAnims: Nie udalo sie zajac miejsca na animacje/kierunek \"%s\"", name.c_str());
				break;
				}

			ani->addDirection(DEFAULT, dir);

			if(parsSpace.count()<7)
				{
				Log::fail("Sprite.loadAnims: Brak wymaganych informacji (\"%s\")", parsSpace.getText().c_str());
				break;
				}

			int count=parsSpace.val(0);
			Frame f( parsSpace.val(1), parsSpace.val(2), parsSpace.val(3), parsSpace.val(4), parsSpace.val(5), parsSpace.val(6) );

			for(int i=0; i<count; ++i)
				{
				dir->addFrame(f);
				f.x+=f.w;
				}
			}
		}

	if(ani)
		spr->addAnimation(ani, aniname);

// Sprawdz czy wczytala sie przynajmniej jedna animacja
	if(spr->getAnims()<1)
		return false;
	for(unsigned i=0u; i<spr->getAnims(); ++i)
		{
		if(spr->getAnim(i).getDir(DEFAULT).getFrames()<1)
			// Dodaj pusta klatke? Chyba sie nie bedzie dalo.
			return false;
		}

	return true;
	}


/**** Wczytywanie xml ****/
Sprite* SpriteManager::loadXml(const std::string& name)
	{
	char *data;

	Sprite *spr;
	TiXmlDocument xml(""); // <- WAZNE! NIE mozna podsylac NULLa do tego konstruktora.

	data=Resource::load(name);
	if(!data)
		{
		Log::fail("SpriteManager.loadXml: Nie udalo sie wczytac xml'a \"%s\"", name.c_str());
		return NULL;
		}

	xml.Parse(data);
	delete [] data;

	TiXmlNode *node=xml.FirstChild(XML_TAG_SPRITE);
	while(node)
		{
		const char *sname;
		sname=node->ToElement()->Attribute(XML_ATR_NAME);
		sname=(sname)?sname:name.c_str();

		try
			{
			spr=new Sprite(name);
			}
		catch(std::bad_alloc&)
			{
			Log::fail("SpriteManager.loadXml: Nie udalo sie zajac miejsca na sprite \"%s\"", name.c_str());
			return NULL;
			}

		xmlParseSprite(spr, node->ToElement());

		if(spr->getAnims()<1)
			{
			Log::debug("SpriteManager.loadXml: Nie dodano zadnej animacji, dodawanie calego obrazka");

			Animation *a=new Animation();
			Direction *d=new Direction();

			d->setReturn(0u);
			d->setSpeed(0.0f);
			d->addFrame( Frame(0, 0, spr->getImg()->w, spr->getImg()->h, 0, 0, 0, 0) );

			a->addDirection(DEFAULT, d);
			spr->addAnimation(a);
			}

		if(sprites.find(sname)!=sprites.end())
			delete sprites[sname];
		sprites[sname]=spr;

		//Log::info("SpriteManager.loadXml: Wczytano sprite \"%s\"", spr->getName().c_str());

		node=xml.IterateChildren(XML_TAG_SPRITE, node);
		}

	if(sprites.find(name)!=sprites.end())
		return sprites[name];
	return NULL;
	}

void SpriteManager::xmlParseSprite(Sprite *spr, TiXmlElement *elem)
	{
	// Wczytanie obrazka
	TiXmlElement *img=elem->FirstChildElement(XML_TAG_IMAGE);

	if(img)
		{
		if(!img->Attribute(XML_ATR_RELATIVE) || spr->getName().length()<1)
			spr->setImage( ImageManager::get(img->GetText()) );
		else
			{
			Parser p(spr->getName(), "/\\:");

			p.del(p.count()-1);
			p.ins(p.count(), img->GetText());

			spr->setImage( ImageManager::get(p.getText()) );
			}
		}
	else
		Log::fail("SpriteManager.xmlParseSprite: Nie ustawiono obrazka");

	// Wczytanie animacji
	TiXmlNode *node=elem->FirstChild(XML_TAG_ANIM);
	while(node)
		{
		const char *name;

		TiXmlElement *e=node->ToElement();

		name=e->Attribute(XML_ATR_NAME);

		//Log::debug("SpriteManager.xmlParseSprite: Animacja, spd %f ret %d", ani.getSpd(), ani.getRet());

		Animation *ani;

		try {
			ani=new Animation();
			}
		catch(std::bad_alloc& e)
			{
			continue;
			}

		//Log::debug("SpriteManaget.xmlParseSprite: anim, name \"%s\"", name);
		xmlParseAnim(ani, e);

		/*unsigned aaa=*/spr->addAnimation(ani, (name)?name:"");
		//Log::debug("SpriteManaget.xmlParseSprite: anim, name \"%s\" %u", name, aaa);

		// Dalsza iteracja
		node=elem->IterateChildren(XML_TAG_ANIM, node);	// Z jakiegos powodu bylo tutaj: node=elem->IterateChildren(XML_TAG_FRAME, node);
		}
	}

void SpriteManager::xmlParseAnim(Animation *ani, TiXmlElement *elem)
	{
	TiXmlNode *node;

	const char *dirstr;
	double spd;
	int ret;

	Directions dirv;

	node=elem->FirstChild(XML_TAG_DIR);
	while(node)
		{
		TiXmlElement *e=node->ToElement();

		spd=1.0f;
		ret=-1;

		dirstr=e->Attribute(XML_ATR_DIRECTION);
		e->Attribute(XML_ATR_SPD, &spd);
		e->Attribute(XML_ATR_RET, &ret);

		if(!dirstr)										dirv=DEFAULT;
		else if(strcmp(dirstr, XML_DIR_RIGHT)==0)		dirv=RIGHT;
		else if(strcmp(dirstr, XML_DIR_RIGHT_DOWN)==0)	dirv=RIGHT_DOWN;
		else if(strcmp(dirstr, XML_DIR_DOWN)==0)		dirv=DOWN;
		else if(strcmp(dirstr, XML_DIR_LEFT_DOWN)==0)	dirv=LEFT_DOWN;
		else if(strcmp(dirstr, XML_DIR_LEFT)==0)		dirv=LEFT;
		else if(strcmp(dirstr, XML_DIR_LEFT_UP)==0)		dirv=LEFT_UP;
		else if(strcmp(dirstr, XML_DIR_UP)==0)			dirv=UP;
		else if(strcmp(dirstr, XML_DIR_RIGHT_UP)==0)	dirv=RIGHT_UP;
		else
			{
			dirv=DEFAULT;

			Log::fail("SpriteManager.xmlParseAnim: Nieprawidlowy kierunek \"%s\"", dirstr);
			}

		Direction *dir;

		try
			{
			dir=new Direction();
			}
		catch(std::bad_alloc&)
			{
			continue;
			}

		dir->setSpeed(spd);
		dir->setReturn(ret);

		//Log::debug("SpriteManaget.xmlParseSprite: dir, spd %f ret %d dir %d (%d)", spd, ret, dirv, (Directions)dirv);
		xmlParseDir(dir, e);

		if(ret<0)
			dir->setReturn(dir->getFrames()-1);

		ani->addDirection(dirv, dir);

		node=elem->IterateChildren(XML_TAG_DIR, node);
		}

	node=elem->FirstChild(XML_TAG_FRAME);
	if(node)
		{
		Direction *dir;

		try
			{
			dir=new Direction();
			}
		catch(std::bad_alloc&)
			{
			return;
			}

		//Log::debug("SpriteManaget.xmlParseSprite: dir, spd %f ret %d dir %d (%d)", spd, ret, dirv, (Directions)dirv);
		xmlParseDir(dir, elem);

		dir->setSpeed(0);
		dir->setReturn(dir->getFrames()-1);

		ani->addDirection(DEFAULT, dir);
		}
	}

void SpriteManager::xmlParseDir(Direction *dir, TiXmlElement *elem)
	{
	TiXmlNode *node;

	node=elem->FirstChild(XML_TAG_FRAME);
	while(node)
		{
		TiXmlElement *e=node->ToElement();

		Frame f(0, 0, 0, 0, 0, 0, 0, 0);

		xmlParseFrame(&f, e);

		//Log::debug("SpriteManaget.xmlParseSprite: frame");
		dir->addFrame(f);

		node=elem->IterateChildren(XML_TAG_FRAME, node);
		}
	}

void SpriteManager::xmlParseFrame(Frame *f, TiXmlElement *elem)
	{
	if(!elem->Attribute(XML_ATR_X,		&f->x)) 	f->x=0;
	if(!elem->Attribute(XML_ATR_Y,		&f->y)) 	f->y=0;
	if(!elem->Attribute(XML_ATR_W,		&f->w)) 	f->w=0;
	if(!elem->Attribute(XML_ATR_H,		&f->h)) 	f->h=0;
	if(!elem->Attribute(XML_ATR_SPOTX,	&f->spotx)) f->spotx=0;
	if(!elem->Attribute(XML_ATR_SPOTY,	&f->spoty)) f->spoty=0;
	if(!elem->Attribute(XML_ATR_ACTX,	&f->actx))  f->actx=0;
	if(!elem->Attribute(XML_ATR_ACTY,	&f->acty))  f->acty=0;
	}
